import gdal
import gdalconst as gc
import pandas as pd
from glob import glob
from dset_config import USE_BASE_PATHS, INPUT_BASE_PATH, OUTPUT_BASE_PATH
from dset_config import STACK_VARS, STACK_VAR_MAP, STACK_VAR_NAME, STACK_VALUE_NAME
from dset_config import ADD_BAND_HOW, DEFAULT_COLUMNS, PRE_B, SILENT



# Main Public Methods
"""PUBLIC Method: tif2df
  open multiband tif as multicolumn dataframe

  columns: (list) column names for (long) dataframe
  stacked: (bool) 
    - if true return in long form: ie stack columns are turned into rows for stack_value_name
    - if false return in multicolumn "short" form
  stack_vars: the column names that are transformed into vars in stacked column
  stack_value_name: (str) name of the value in the stack columns
  stack_var_name: (str) name of column that contains the long-column-names
  dropna: (str or bool) Drop NaNs: if True dropna(), else dropna(how=dropna) default=True. 
  zerona: (bool) Set NaNs to zero. defalut=False - a True value will override dropna
  addindex: (bool) Add column of index values. 
    * Note: this column is always added for "stacked" datasets
  is_full_path: (bool) see open_raster method

  returns dataframe 
"""
def tif2df(path,columns=None,stacked=True,stack_vars=None,stack_value_name=None,stack_var_name=None,column_map=None,dropna=True,zerona=False,addindex=False,add_band_how=ADD_BAND_HOW,is_full_path=False):
  ds=open_raster(path,is_full_path)
  df=None
  for b in range(1,ds.RasterCount+1): 
    if not SILENT: print "\t{}...".format(b)
    df=_add_band(df,band_df(ds,b),add_band_how)
  return _postprocess(df,columns,stacked,stack_vars,stack_value_name,stack_var_name,column_map,dropna,zerona,addindex)



"""PUBLIC Method: tifs2df
  open multiple single band tifs as a single multicolumn dataframe

  columns: (list) column names for (long) dataframe
  stacked: (bool) 
    - if true return in long form: ie stack columns are turned into rows for stack_value_name
    - if false return in multicolumn "short" form
  stack_vars: the column names that are transformed into vars in stacked column
  stack_value_name: (str) name of the value in the stack columns
  stack_var_name: (str) name of column that contains the long-column-names
  dropna: (str or bool) Drop NaNs: if True dropna(), else dropna(how=dropna) default=True. 
  zerona: (bool) Set NaNs to zero. defalut=False - a True value will override dropna
  addindex: (bool) Add column of index values. 
    * Note: this column is always added for "stacked" datasets
  is_full_path: (bool) see open_raster method

  returns dataframe 
"""
def tifs2df(paths_or_prefix,columns=None,stacked=True,stack_vars=None,stack_value_name=None,stack_var_name=None,column_map=None,dropna=True,zerona=False,addindex=False,add_band_how=ADD_BAND_HOW,is_full_path=False):
  if type(paths_or_prefix) is str: files=glob("{}*".format(paths_or_prefix))
  else: files=paths_or_prefix
  df=None
  b=1
  for file in files:
    if not SILENT: print "\t{}...".format(b)
    ds=open_raster(file,is_full_path)
    df=_add_band(df,band_df(ds,1,"{}{}".format(PRE_B,b)),add_band_how)
    b=b+1 
  return _postprocess(df,columns,stacked,stack_vars,stack_value_name,stack_var_name,column_map,dropna,zerona,addindex)




# Additional Public Methods
"""PUBLIC Method: open_raster
  reads raster file into gdal dataset

  path: (str) path to file
  is_full_path: (bool) defaults to true
    - if true, open the file at path provided
    - if false, append the config.py's INPUT_BASE_PATH to the path provided.
  
  returns gdal dataset   
"""
def open_raster(path, is_full_path=False):
  if USE_BASE_PATHS and not is_full_path: 
    path="{}/{}".format(INPUT_BASE_PATH,path)
  return gdal.Open(path, gc.GA_ReadOnly)



"""PUBLIC Method: band_df
  turns specified band of gdal dataset into a pandas dataframe

  ds: (gdal dataset)
  band_index: index of band to return as dataframe. defaults to 1
  column: column name for dataframe
  size: number of rows (will be calculated if not provided)
  
  returns dataframe  
"""
def band_df(ds,band_index=1,column=None,size=None):
  if not size: size=ds.RasterXSize * ds.RasterYSize
  return pd.DataFrame(ds.GetRasterBand(band_index).ReadAsArray().reshape(size,1),columns=[column or "{}{}".format(PRE_B,band_index)])




# Internal Methods
"""PRIVATE Method: _add_band
  safe add band to dataframe

  df: (dataframe) inital dataframe (or None if first)
  band: (dataframe) single column dataframe to add
  returns dataframe  
"""
def _add_band(df,band,how=ADD_BAND_HOW):
  if df is None: df=band
  else: df=df.join(band,how=how)
  return df


"""PRIVATE Method: _postprocess
  common transformations applied to dataframe after it is created in tif2df and tifs2df

  properties: see public methods

  returns dataframe  
"""
def _postprocess(df,columns,stacked,stack_vars,stack_value_name,stack_var_name,column_map,dropna,zerona,addindex):
  if zerona: df=df.fillna(0)
  elif dropna:
    if dropna is True: df=df.dropna().copy()
    else: df=df.dropna(how=dropna).copy()
  if columns is not None: df.columns=columns
  elif DEFAULT_COLUMNS: df.columns=DEFAULT_COLUMNS
  if stacked: 
    if not stack_vars: 
      if STACK_VARS: stack_vars=STACK_VARS
      else: stack_vars=columns
    df=_stack_year_bands(df,stack_value_name,stack_var_name,stack_vars,column_map)
  elif addindex: df['index']=df.index
  return df


"""PRIVATE Method: _postprocess
  common transformations applied to dataframe after it is created in tif2df and tifs2df

  properties: see public methods

  returns dataframe  
"""
def _stack_year_bands(df,stack_value_name,stack_var_name,stack_vars=False,stack_var_map=False):
  if not stack_var_map: stack_var_map=STACK_VAR_MAP
  if not stack_value_name: 
    if STACK_VALUE_NAME: stack_value_name=STACK_VALUE_NAME
    else: stack_value_name="value"
  if not stack_var_name: 
    if STACK_VAR_NAME: stack_var_name=STACK_VAR_NAME
    else: stack_var_name="stack"
  df['index']=df.index
  df=pd.melt(df,id_vars=['index'],value_vars=stack_vars,var_name=stack_var_name,value_name=stack_value_name)
  if stack_var_map: df[stack_var_name]=df[stack_var_name].apply(lambda band: stack_var_map[band])
  return df


